@@ -25,14 +25,34 @@ import logging
 logger = logging.getLogger('resultsdb_api')
 logger.addHandler(logging.NullHandler())
 
+_KEEP = object()
 
-def _fname():
+
+def _fparams(expand_kwargs=True):
+    """Gets the parameters of the function, from which _fparams is called
+    and returns the list of params, minus `self`"""
     frame = inspect.currentframe().f_back
-    name = frame.f_code.co_name
-    return name
+    args, varargs, keywords, values = inspect.getargvalues(frame)
+
+    params = {}
+
+    for key in args:
+        if key == 'self':
+            continue
+        params[key] = values[key]
+
+    if keywords:
+        if expand_kwargs:
+            for key, value in values[keywords].iteritems():
+                params[key] = value
+        else:
+            params[keywords] = values[keywords]
+
+    return params
 
 
 class ResultsDBapiException(Exception):
+
     def __init__(self, message='', response=None):
         ''':param response: :class:`requests.Response` object'''
         self.message = message
@@ -43,8 +63,12 @@ class ResultsDBapiException(Exception):
 
 
 class ResultsDBapi(object):
+
     def __init__(self, api_url):
-        self.url = api_url
+        # remove trailing slash(es), so we don't generate
+        # urls with a double slash which breaks werkzeug
+        # https://github.com/mitsuhiko/werkzeug/issues/491
+        self.url = api_url.rstrip('/')
 
     def __raise_on_error(self, r):
         if r.ok:
@@ -62,157 +86,125 @@ class ResultsDBapi(object):
         except KeyError:
             raise ResultsDBapiException('HTTP %s Error' % r.status_code, r)
 
+    def __prepare_params(self, params_all):
+        params = {}
+        for key, value in params_all.iteritems():
+            if value is None:
+                continue
+            if key == 'raw_params':
+                continue
 
-    def create_job(self, ref_url, status = None, name = None, uuid = None):
-        data = dict(ref_url = unicode(ref_url), status = status, name = name, uuid = uuid)
+            # if a param's name ends with _like, we treat it as if :like filter should be applied
+            #  for the rare case, where it really is supposed to be the name, user should provide
+            #  it in the 'raw_params' dict
+            if key.endswith('_like'):
+                key = "%s:like" % key[:len('_like')]
 
-        url = "%s/jobs" % self.url
-        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
-        r = requests.post(url, data=json.dumps(data), headers = headers)
-        self.__raise_on_error(r)
-
-        return r.json()
-
-    def update_job(self, id = None, href = None, status = None, return_data = False):
-        """
-        id or href required, href has preference if both are set
-        """
-        url = None
-        if id:
-            url = "%s/jobs/%s" % (self.url, id)
-        if href:
-            url = href
-        if not url:
-            raise TypeError("%s() requires either id or href parameter" % _fname())
-
-        data = dict(status = status, return_data = return_data)
-
-        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
-        r = requests.put(url, data=json.dumps(data), headers = headers)
-        self.__raise_on_error(r)
-
-        return r.json()
-
-
-    def get_job(self, id = None, href = None):
-        url = None
-        if id:
-            url = "%s/jobs/%s" % (self.url, id)
-        if href:
-            url = href
-        if not url:
-            raise TypeError("%s() requires either id or href parameter" % _fname())
-
-        r = requests.get(url)
-        self.__raise_on_error(r)
-
-        return r.json()
-
-    def get_jobs(self, page = None, limit = None, since = None, status = None, name = None, load_results = False, **kwargs):
-        url = "%s/jobs" % self.url
-
-        params_all = dict(page = page, limit = limit, since = since, status = status, name = name)
-        params = {key:value for key, value in params_all.iteritems() if value is not None}
-        for key, value in params.iteritems():
-            if type(value) in (list, tuple):
-                params[key] = ','.join([unicode(v) for v in value])
-            else:
-                params[key] = unicode(value)
-        if load_results is True:
-            params['load_results'] = 'true'
-        r = requests.get(url, params = params)
-        self.__raise_on_error(r)
-
-        r = r.json()
-        return r
-
-
-    def create_result(self, job_id, testcase_name, outcome, summary = None, log_url = None, **kwargs):
-        url = "%s/results" % self.url
-
-        data = dict(job_id = job_id,
-                    testcase_name = testcase_name,
-                    outcome = outcome,
-                    summary = summary,
-                    log_url = log_url,
-                    result_data = kwargs)
-
-        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
-        r = requests.post(url, data=json.dumps(data), headers = headers)
-        self.__raise_on_error(r)
-
-        return r.json()
-
-    def get_result(self, id = None, href = None):
-        url = None
-        if id:
-            url = "%s/results/%s" % (self.url, id)
-        if href:
-            url = href
-        if not url:
-            raise TypeError("%s() requires either id or href parameter" % _fname())
-
-        r = requests.get(url)
-        self.__raise_on_error(r)
-
-        return r.json()
-
-    def get_results(self, page = None, limit = None, since = None, outcome = None, job_id = None, testcase_name = None, **kwargs):
-        url = "%s/results" % self.url
-        params = dict(page = page, limit = limit, since = since, outcome = outcome, job_id = job_id, testcase_name = testcase_name)
-        params.update(kwargs)
-        params = {key:value for key, value in params.iteritems() if value is not None}
-        for key, value in params.iteritems():
             if type(value) in (list, tuple):
                 params[key] = ','.join([unicode(v) for v in value])
             else:
                 params[key] = unicode(value)
 
+        if 'raw_params' in params_all.keys() and params_all['raw_params']:
+            raw_params = {
+                key: unicode(value) for key, value in params_all['raw_params'].iteritems()}
+            params.update(raw_params)
+        return params
 
-        r = requests.get(url, params = params)
-        self.__raise_on_error(r)
-
-        r = r.json()
-        return r
-
-    def create_testcase(self, name, url):
-
-        data = dict(name = name, url = url)
+    def create_group(self, uuid=None, ref_url=None, description=None):
+        url = "%s/groups" % self.url
         headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
-
-        r = requests.post("%s/testcases" % self.url , data = json.dumps(data), headers = headers)
+        r = requests.post(url, data=json.dumps(_fparams()), headers=headers)
         self.__raise_on_error(r)
 
         return r.json()
 
+    def update_group(self, uuid, ref_url=_KEEP, description=_KEEP):
+        data = {}
+        if ref_url is not _KEEP:
+            data['ref_url'] = ref_url
+        if description is not _KEEP:
+            data['description'] = description
+        if data:
+            data['uuid'] = uuid
 
-    def update_testcase(self, name, url):
-        data = dict(url = url)
-
+        url = "%s/groups" % self.url
         headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
-        rurl = "%s/testcases/%s" % (self.url, name)
-        r = requests.put(rurl, data = json.dumps(data), headers = headers)
+        r = requests.post(url, data=json.dumps(data), headers=headers)
+        self.__raise_on_error(r)
+
+        return r.json()
+
+    def get_group(self, uuid):
+        url = "%s/groups/%s" % (self.url, uuid)
+        r = requests.get(url)
+        self.__raise_on_error(r)
+
+        return r.json()
+
+    def get_groups(self, page=None, limit=None, description=None, description_like=None, uuid=None):
+        url = "%s/groups" % self.url
+        r = requests.get(url, params=self.__prepare_params(_fparams()))
+        self.__raise_on_error(r)
+
+        return r.json()
+
+    def create_result(self, outcome, testcase, groups=None, note=None, ref_url=None, **data):
+        url = "%s/results" % self.url
+        data = _fparams(expand_kwargs=False)
+        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
+        r = requests.post(url, data=json.dumps(data), headers=headers)
+        self.__raise_on_error(r)
+
+        return r.json()
+
+    def get_result(self, id):
+        url = "%s/results/%s" % (self.url, id)
+        r = requests.get(url)
+        self.__raise_on_error(r)
+
+        return r.json()
+
+    def get_results(self, page=None, limit=None, since=None, outcome=None, groups=None, testcases=None, testcases_like=None, raw_params=None, **kwargs):
+        url = "%s/results" % self.url
+        r = requests.get(url, params=self.__prepare_params(_fparams()))
+        self.__raise_on_error(r)
+
+        return r.json()
+
+    def create_testcase(self, name, ref_url=None):
+        url = "%s/testcases" % self.url
+        data = _fparams()
+        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
+        r = requests.post(url, data=json.dumps(data), headers=headers)
+        self.__raise_on_error(r)
+
+        return r.json()
+
+    def update_testcase(self, name, url=_KEEP):
+        data = {}
+        if ref_url is not _KEEP:
+            data['ref_url'] = ref_url
+        if data:
+            data['name'] = name
+
+        url = "%s/testcases" % self.url
+        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
+        r = requests.post(url, data=json.dumps(data), headers=headers)
         self.__raise_on_error(r)
 
         return r.json()
 
     def get_testcase(self, name):
         url = "%s/testcases/%s" % (self.url, name)
-
         r = requests.get(url)
         self.__raise_on_error(r)
 
         return r.json()
 
-    def get_testcases(self, page = None, limit = None, **kwargs):
+    def get_testcases(self, page=None, limit=None, name=None, name_like=None):
         url = "%s/testcases" % self.url
-
-        params = dict(page = page, limit = limit)
-        params = {key:value for key, value in params.iteritems() if value is not None}
-
-        r = requests.get(url, params = params)
+        r = requests.get(url, params=self.__prepare_params(_fparams()))
         self.__raise_on_error(r)
 
-        r = r.json()
-        return r
-
+        return r.json()
